
import Queue
from threading import Thread
import logging
import time
import os.path

from posttroll.publisher import Publish
from posttroll.message import Message
from trollflow_sat import utils
from trollsift import compose


class DataWriterContainer(object):

    '''Container for DataWriter instance
    '''

    logger = logging.getLogger("DataWriterContainer")
    _prev_lock = None

    def __init__(self, topic=None, port=0, nameservers=None,
                 save_settings=None, use_lock=False):
        self.topic = topic
        self._input_queue = None
        self.output_queue = None  # Queue.Queue()
        self.thread = None
        self.use_lock = use_lock

        if nameservers is None:
            nameservers = []
        else:
            if type(nameservers) not in (list, tuple, set):
                nameservers = [nameservers, ]

        # Create a Writer instance
        self.writer = DataWriter(queue=self.input_queue,
                                 save_settings=save_settings,
                                 topic=topic,
                                 port=port, nameservers=nameservers,
                                 prev_lock=self._prev_lock)
        # Start Writer instance into a new daemonized thread.
        self.thread = Thread(target=self.writer.run)
        self.thread.setDaemon(True)
        self.thread.start()

    @property
    def input_queue(self):
        """Property writer"""
        return self._input_queue

    @input_queue.setter
    def input_queue(self, queue):
        """Set writer queue"""
        self._input_queue = queue
        self.writer.queue = queue

    @property
    def prev_lock(self):
        """Property writer"""
        return self._prev_lock

    @prev_lock.setter
    def prev_lock(self, lock):
        """Set lock of the previous worker"""
        if self.use_lock:
            self._prev_lock = lock
            self.writer.prev_lock = lock

    def __setstate__(self, state):
        self.__init__(**state)

    def restart(self):
        '''Restart writer after configuration update.
        '''
        if self.writer is not None:
            if self.writer.loop:
                self.stop()
        self.__init__()

    def stop(self):
        '''Stop writer.'''
        self.logger.debug("Stopping writer.")
        self.writer.stop()
        self.thread.join()
        self.logger.debug("Writer stopped.")
        self.thread = None

    def is_alive(self):
        """Return the thread status"""
        return self.thread.is_alive()


class DataWriter(Thread):

    """Writes data to disk.
    """

    logger = logging.getLogger("DataWriter")

    def __init__(self, queue=None, save_settings=None,
                 topic=None, port=0, nameservers=None, prev_lock=None):
        Thread.__init__(self)
        self.queue = queue
        self._loop = False
        if save_settings is not None:
            self._save_settings = save_settings
        else:
            self._save_settings = {}
        self._port = port
        if nameservers is None:
            self._nameservers = []
        else:
            if type(nameservers) not in (list, tuple, set):
                nameservers = [nameservers, ]
            self._nameservers = nameservers
        self._topic = topic
        self.prev_lock = prev_lock

    def run(self):
        """Run the thread."""
        self._loop = True
        # Parse settings for saving
        compression = self._save_settings.get('compression', 6)
        tags = self._save_settings.get('tags', None)
        fformat = self._save_settings.get('fformat', None)
        gdal_options = self._save_settings.get('gdal_options', None)
        blocksize = self._save_settings.get('blocksize', None)

        # Initialize publisher context
        with Publish("l2producer", port=self._port,
                     nameservers=self._nameservers) as pub:

            while self._loop:
                if self.queue is not None:
                    try:
                        obj = self.queue.get(True, 1)
                        if self.prev_lock is not None:
                            self.logger.debug("Writer acquires lock of "
                                              "previous worker: %s",
                                              str(self.prev_lock))
                            utils.acquire_lock(self.prev_lock)
                        self.queue.task_done()
                    except Queue.Empty:
                        continue
                    for fname in obj.info["fnames"]:
                        self.logger.info("Saving %s", fname)
                        obj.save(fname, compression=compression, tags=tags,
                                 fformat=fformat, gdal_options=gdal_options,
                                 blocksize=blocksize)

                        area = getattr(obj, "area")
                        try:
                            area_data = {"name": area.name,
                                         "area_id": area.area_id,
                                         "proj_id": area.proj_id,
                                         "proj4": area.proj4_string,
                                         "shape": (area.x_size, area.y_size)
                                         }
                        except AttributeError:
                            area_data = None

                        to_send = {"nominal_time": getattr(obj, "time_slot"),
                                   "uid": os.path.basename(fname),
                                   "uri": os.path.abspath(fname),
                                   "area": area_data,
                                   "productname": obj.info["productname"]
                                   }
                        # if self._topic is not None:
                        if self._topic is not None:
                            topic = self._topic
                            if area_data is not None:
                                topic = compose(topic,  area_data)
                            else:
                                topic = compose(topic, {'area_id': 'satproj'})
                            msg = Message(topic, "file", to_send)
                            pub.send(str(msg))
                            self.logger.debug("Sent message: %s", str(msg))
                        self.logger.info("Saved %s", fname)

                    del obj
                    obj = None
                    # After all the items have been processed, release the
                    # lock for the previous worker
                    if self.prev_lock is not None:
                        utils.release_locks([self.prev_lock],
                                            self.logger.debug,
                                            "Writer releses lock of "
                                            "previous worker: %s" %
                                            str(self.prev_lock))
                else:
                    time.sleep(1)

    def stop(self):
        """Stop writer."""
        self._loop = False

    @property
    def loop(self):
        """Property loop"""
        return self._loop



"""Classes for handling area gathering for Trollflow based Trollduction"""

import logging
import Queue
from threading import Thread
import time
import datetime as dt

from posttroll import message

from trollduction.collectors import region_collector


class AreaGathererContainer(object):

    """Container for area gatherer."""

    logger = logging.getLogger("AreaGathererContainer")

    def __init__(self, config):
        self.gatherer = None
        self._input_queue = None
        self.output_queue = Queue.Queue()
        self.thread = None

        # Create a AreaGatherer instance
        self.gatherer = AreaGatherer(config, self.input_queue,
                                     self.output_queue)

        # Start AreaGatherer into a new daemonized thread.
        self.thread = Thread(target=self.gatherer.run)
        self.thread.setDaemon(True)
        self.thread.start()

    @property
    def input_queue(self):
        """Input queue property"""
        return self._input_queue

    @input_queue.setter
    def input_queue(self, queue):
        """Setter for input queue property"""
        self._input_queue = queue
        self.gatherer.input_queue = queue

    def __setstate__(self, state):
        self.__init__(**state)

    def restart(self):
        '''Restart writer after configuration update.
        '''
        if self.gatherer is not None:
            if self.gatherer.loop:
                self.stop()
        self.__init__()

    def stop(self):
        """Stop gatherer."""
        self.logger.debug("Stopping AreaGatherer.")
        self.gatherer.stop()
        self.thread.join()
        self.logger.debug("AreaGatherer stopped.")
        self.thread = None

    def is_alive(self):
        """Return the thread status"""
        return self.thread.is_alive()


class AreaGatherer(Thread):

    """Class for gathering segments together to cover a given area."""

    logger = logging.getLogger("AreaGatherer")

    def __init__(self, config, input_queue, output_queue):
        Thread.__init__(self)
        self.input_queue = input_queue
        self.output_queue = output_queue
        self._loop = False

        self.timeliness = dt.timedelta(minutes=config["timeliness"])
        self._pattern = config["pattern"]

        try:
            self._min_coverage = config["min_coverage"] / 100.0
        except KeyError:
            self._min_coverage = 0.0
        try:
            self._duration = dt.timedelta(seconds=config["duration"])
        except KeyError:
            self._duration = None

        if config["aliases"] is not None:
            self._aliases = config["aliases"]
        else:
            self._aliases = {}
        if config["metadata"] is not None:
            self._metadata = config["metadata"]
        else:
            self._metadata = {}

        self._platform_names = config["platform_names"]
        self._regions = config["regions"]

        self.collectors = {}
        self.create_collectors()

    def create_collectors(self):
        """Create area collectors for each platform."""
        for platform in self._platform_names:
            self.collectors[platform] = \
                [region_collector.RegionCollector(region,
                                                  self._timeliness,
                                                  self._duration) for
                 region in self._regions]

    @property
    def loop(self):
        """Loop property"""
        return self._loop

    def _check_timeouts(self):
        """Check if timeouts have happened."""
        for platform in self._platforms:
            timeouts = [(collector, collector.timeout)
                        for collector in self.collectors[platform]
                        if collector.timeout is not None]
            if timeouts:
                next_timeout = min(timeouts, key=(lambda x: x[1]))
                if (next_timeout[1] and
                        (next_timeout[1] < dt.datetime.utcnow())):
                    self.logger.warning("Timeout detected, terminating "
                                        "collector")
                    self.logger.debug("Area: %s, timeout: %s",
                                      next_timeout[0].region,
                                      str(next_timeout[1]))
                    self.terminator(next_timeout[0].finish())
                else:
                    self.logger.debug("Waiting %s seconds until timeout",
                                      str(total_seconds(next_timeout[1] -
                                                        dt.datetime.utcnow())))

    def terminator(self, metadata):
        """Dummy terminator function.
        """
        sorted_mda = sorted(metadata, key=lambda x: x["start_time"])

        mda = metadata[0].copy()

        mda['end_time'] = sorted_mda[-1]['end_time']
        mda['collection_area_id'] = sorted_mda[-1]['collection_area_id']
        mda['collection'] = []

        is_correct = False
        for meta in sorted_mda:
            new_mda = {}
            if "uri" in meta or 'dataset' in meta:
                is_correct = True
            for key in ['dataset', 'uri', 'uid']:
                if key in meta:
                    new_mda[key] = meta[key]
                new_mda['start_time'] = meta['start_time']
                new_mda['end_time'] = meta['end_time']
            mda['collection'].append(new_mda)

        for key in ['dataset', 'uri', 'uid']:
            if key in mda:
                del mda[key]

        if is_correct:
            msg = message.Message("/placeholder", "collection",
                                  mda)
            self.logger.info("Collection ready %s", str(msg))
            self.output_queue.put(msg)
        else:
            self.logger.warning("Malformed metadata, no key: %s", "uri")

    def run(self):
        """Run SegmentGatherer"""
        self._loop = True
        while self._loop:
            msg = None
            if self.input_queue is not None:
                # Check timeouts
                self._check_timeouts()

                try:
                    # Get new message from the queue
                    msg = self.input_queue.get(True, 1)
                    metadata = msg.data
                    # Replace aliases
                    for key, aliases in self._aliases.items():
                        if key in metadata:
                            metadata[key] = aliases.get(metadata[key],
                                                        metadata[key])
                    for collector in self.collectors["platform_name"]:
                        res = collector(metadata)
                        if res:
                            msg = self.terminator(res)
                            self.output_queue.put(msg)
                except KeyboardInterrupt:
                    self.stop()
                    continue
                except Queue.Empty:
                    continue
            else:
                time.sleep(1)
                continue

    def stop(self):
        """Stop gatherer."""
        self.logger.info("Stopping AreaGatherer.")
        self._loop = False


def total_seconds(tdef):
    """Calculate total time in seconds.
    """
    return ((tdef.microseconds +
             (tdef.seconds + tdef.days * 24 * 3600) * 10 ** 6) / 10.0 ** 6)



"""Classes for handling segment gathering for Trollflow based Trollduction"""

import logging
import Queue
from threading import Thread
import time
import datetime as dt
import yaml
from collections import OrderedDict

from trollsift import Parser
from posttroll import message

SLOT_NOT_READY = 0
SLOT_READY = 1
SLOT_READY_BUT_WAIT_FOR_MORE = 2
SLOT_OBSOLETE_TIMEOUT = 3

DO_NOT_COPY_KEYS = ("uid", "uri", "channel_name", "segment")


class SegmentGathererContainer(object):

    """Container for SegmentGatherer instance"""

    logger = logging.getLogger("SegmentGathererContainer")

    def __init__(self, config):
        self.gatherer = None
        self._input_queue = None
        self.output_queue = Queue.Queue()
        self.thread = None

        # Create a SegmentGatherer instance
        self.gatherer = SegmentGatherer(config, self.input_queue,
                                        self.output_queue)

        # Start SegmentGatherer into a new daemonized thread.
        self.thread = Thread(target=self.gatherer.run)
        self.thread.setDaemon(True)
        self.thread.start()

    @property
    def input_queue(self):
        """Input queue property"""
        return self._input_queue

    @input_queue.setter
    def input_queue(self, queue):
        """Setter for input queue property"""
        self._input_queue = queue
        self.gatherer.input_queue = queue

    def __setstate__(self, state):
        self.__init__(**state)

    def restart(self):
        '''Restart writer after configuration update.
        '''
        if self.gatherer is not None:
            if self.gatherer.loop:
                self.stop()
        self.__init__()

    def stop(self):
        """Stop gatherer."""
        self.logger.debug("Stopping SegmentGatherer.")
        self.gatherer.stop()
        self.thread.join()
        self.logger.debug("SegmentGatherer stopped.")
        self.thread = None

    def is_alive(self):
        """Return the thread status"""
        return self.thread.is_alive()


class SegmentGatherer(Thread):

    """Class for gathering segments of one time slot together."""

    logger = logging.getLogger("SegmentGatherer")

    def __init__(self, config, input_queue, output_queue):
        Thread.__init__(self)
        self.input_queue = input_queue
        self.output_queue = output_queue
        self._loop = False
        self._parsers = []
        with open(config, 'r') as fid:
            self._config = yaml.load(fid)

        self._set_parsers()

        try:
            self._timeliness = \
                dt.timedelta(seconds=self._config["config"]["timeliness"])
        except KeyError:
            self._timeliness = dt.timedelta(seconds=1200)
        try:
            self._num_files_premature_publish = \
                self._config["config"]["num_files_premature_publish"]
        except KeyError:
            self._num_files_premature_publish = -1

        self.slots = OrderedDict()

        self.time_name = self._config["config"]["time_name"]

    def _clear_data(self, time_slot):
        """Clear data."""
        if time_slot in self.slots:
            del self.slots[time_slot]

    def _init_data(self, msg, mda):
        """Init wanted, all and critical files"""
        # Init metadata struct
        metadata = {}
        for key in msg.data:
            if key not in DO_NOT_COPY_KEYS:
                metadata[key] = msg.data[key]
        metadata['dataset'] = []

        # Use also metadata parsed from the filenames
        metadata.update(mda)

        time_slot = str(metadata[self.time_name])
        self.slots[time_slot] = {}
        self.slots[time_slot]['metadata'] = metadata.copy()

        # Critical files that are required, otherwise production will fail.
        # If there are no critical files, empty set([]) is used.
        # self.slots[time_slot]['critical_files'] = \
        self._compose_filenames(time_slot)

        self.slots[time_slot]['received_files'] = set([])
        self.slots[time_slot]['delayed_files'] = dict()
        self.slots[time_slot]['missing_files'] = set([])
        self.slots[time_slot]['timeout'] = None
        self.slots[time_slot]['files_till_premature_publish'] = \
            self._num_files_premature_publish

    def _set_parsers(self):
        """Set parsers"""
        files = self._config["files"]

        for fle in files:
            pattern = fle["pattern"]
            parser = Parser(pattern)
            self._parsers.append(parser)

    def _compose_filenames(self, time_slot):
        """Compose filename set()s"""

        # Get copy of metadata
        meta = self.slots[time_slot]['metadata'].copy()

        # Replace variable tags (such as processing time) with
        # wildcards, as these can't be forecasted.
        try:
            ignored_keys = self._config["config"]["variable_tags"].split(',')
            meta = _copy_without_ignore_items(meta, ignored_keys=ignored_keys)
        except KeyError:
            pass

        critical_files, wanted_files, all_files = [], [], []
        files = self._config["files"]

        for fle in files:
            pattern = fle["pattern"]
            parser = Parser(pattern)
            for seg in fle["segments"]:
                chans = seg.get('channel_name', [''])
                critical_segments = seg.get('critical_segments', [''])
                wanted_segments = seg.get('wanted_segments', [''])
                all_segments = seg.get('all_segments', [''])
                for chan in chans:
                    meta['channel_name'] = chan
                    for seg2 in critical_segments:
                        meta['segment'] = seg2
                        critical_files.append(parser.globify(meta))
                    for seg2 in wanted_segments:
                        meta['segment'] = seg2
                        wanted_files.append(parser.globify(meta))
                    for seg2 in all_segments:
                        meta['segment'] = seg2
                        all_files.append(parser.globify(meta))

        self.slots[time_slot]['critical_files'] = set(critical_files)
        self.slots[time_slot]['wanted_files'] = set(wanted_files)
        self.slots[time_slot]['all_files'] = set(all_files)

    def slot_ready(self, slot):
        """Determine if slot is ready to be published."""
        # If no files have been collected, return False
        if len(slot['received_files']) == 0:
            return SLOT_NOT_READY

        time_slot = str(slot['metadata'][self.time_name])

        wanted_and_critical_files = \
            slot['wanted_files'].union(slot['critical_files'])
        num_wanted_and_critical_files_received = \
            len(wanted_and_critical_files & slot['received_files'])

        # self.logger.debug("Got %d wanted or critical files in slot %s.",
        #                   num_wanted_and_critical_files_received,
        #                   time_slot)

        if num_wanted_and_critical_files_received \
                == slot['files_till_premature_publish']:
            slot['files_till_premature_publish'] = -1
            return SLOT_READY_BUT_WAIT_FOR_MORE

        # If all wanted files have been received, return True
        if wanted_and_critical_files.issubset(
                slot['received_files']):
            self.logger.info("All files received for slot %s.",
                             time_slot)
            return SLOT_READY

        if slot['critical_files'].issubset(slot['received_files']):
            # All critical files have been received
            if slot['timeout'] is None:
                # Set timeout
                slot['timeout'] = dt.datetime.utcnow() + self._timeliness
                self.logger.info("Setting timeout to %s for slot %s.",
                                 str(slot['timeout']),
                                 time_slot)
                return SLOT_NOT_READY
            elif slot['timeout'] < dt.datetime.utcnow():
                # Timeout reached, collection ready
                self.logger.info("Timeout occured, required files received "
                                 "for slot %s.", time_slot)
                return SLOT_READY
            else:
                pass
        else:
            if slot['timeout'] is None:
                slot['timeout'] = dt.datetime.utcnow() + self._timeliness
                self.logger.info("Setting timeout to %s for slot %s",
                                 str(slot['timeout']),
                                 time_slot)
                return SLOT_NOT_READY

            elif slot['timeout'] < dt.datetime.utcnow():
                # Timeout reached, collection is obsolete
                self.logger.warning("Timeout occured and required files "
                                    "were not present, data discarded for "
                                    "slot %s.",
                                    time_slot)
                return SLOT_OBSOLETE_TIMEOUT
            else:
                pass

        # Timeout not reached, wait for more files
        return SLOT_NOT_READY

    @property
    def loop(self):
        """Loop property"""
        return self._loop

    def run(self):
        """Run SegmentGatherer"""
        self._loop = True
        while self._loop:
            # Check if there are slots ready for publication
            slots = self.slots.copy()
            for slot in slots:
                slot = str(slot)
                status = self.slot_ready(slots[slot])
                if status == SLOT_READY:
                    # Collection ready, publish and remove
                    self._publish(slot)
                    self._clear_data(slot)
                if status == SLOT_READY_BUT_WAIT_FOR_MORE:
                    # Collection ready, publish and but wait for more
                    self._publish(slot, missing_files_check=False)
                elif status == SLOT_OBSOLETE_TIMEOUT:
                    # Collection unfinished and obslote, discard
                    self._clear_data(slot)
                else:
                    # Collection unfinished, wait for more data
                    pass

            # Check queue for new data
            msg = None
            if self.input_queue is not None:
                try:
                    msg = self.input_queue.get(True, 1)
                except KeyboardInterrupt:
                    self.stop()
                    continue
                except Queue.Empty:
                    continue
            else:
                time.sleep(1)
                continue

            if msg.type == "file":
                self.logger.info("New message received: %s", str(msg))
                self.process(msg)

    def stop(self):
        """Stop gatherer."""
        self.logger.info("Stopping gatherer.")
        self._loop = False

    def process(self, msg):
        """Process message"""

        mda = None
        for parser in self._parsers:
            try:
                mda = parser.parse(msg.data["uid"])
                break
            except ValueError:
                continue

        if mda is None:
            self.logger.debug("Unknown file, skipping.")
            return

        metadata = {}
        for key in msg.data:
            if key not in DO_NOT_COPY_KEYS:
                metadata[key] = msg.data[key]
        metadata.update(mda)

        time_slot = str(metadata[self.time_name])

        # Init metadata etc if this is the first file
        if time_slot not in self.slots:
            self._init_data(msg, mda)

        slot = self.slots[time_slot]

        # Replace variable tags (such as processing time) with
        # wildcards, as these can't be forecasted.
        try:
            variable_tags = self._config["config"]["variable_tags"].split(',')
            mda = _copy_without_ignore_items(mda, ignored_keys=variable_tags)
        except KeyError:
            pass

        mask = None
        for parser in self._parsers:
            try:
                mask = parser.globify(mda)
                break
            except ValueError:
                continue

        if mask is None or mask in slot['received_files']:
            return

        # Add uid and uri
        slot['metadata']['dataset'].append({'uri': msg.data['uri'],
                                            'uid': msg.data['uid']})

        # If critical files have been received but the slot is
        # not complete, add the file to list of delayed files
        if len(slot['critical_files']) > 0 and \
           slot['critical_files'].issubset(slot['received_files']):
            delay = dt.datetime.utcnow() - (slot['timeout'] - self._timeliness)
            slot['delayed_files'][msg.data['uid']] = delay.total_seconds()

        # Add to received files
        slot['received_files'].add(mask)

    def _publish(self, time_slot, missing_files_check=True):
        """Publish file dataset and reinitialize gatherer."""

        data = self.slots[time_slot]

        # Diagnostic logging about delayed ...
        delayed_files = data['delayed_files']
        if len(delayed_files) > 0:
            file_str = ''
            for key in delayed_files:
                file_str += "%s %f seconds, " % (key, delayed_files[key])
            self.logger.warning("Files received late: %s",
                                file_str.strip(', '))

        if missing_files_check:
            # and missing files
            missing_files = \
                data['all_files'].difference(data['received_files'])
            if len(missing_files) > 0:
                self.logger.warning("Missing files: %s",
                                    ', '.join(missing_files))

        # Although we're not publishing a message, generate one anyway
        # for compatibility
        msg = message.Message("/placeholder", "dataset", data['metadata'])
        self.logger.info("Forwarding: %s", str(msg))
        self.output_queue.put(msg)


def _copy_without_ignore_items(the_dict, ignored_keys=['ignore']):
    """
    get a copy of *the_dict* without entries having substring
    'ignore' in key
    """
    new_dict = {}
    for (key, val) in list(the_dict.items()):
        if key not in ignored_keys:
            new_dict[key] = val
    return new_dict



"""Class for handling area coverage checks for Trollflow based Trollduction"""

import logging
import time

from trollflow.workflow_component import AbstractWorkflowComponent
from trollflow_sat import utils
from trollsched.satpass import Pass


class CoverageCheck(AbstractWorkflowComponent):

    """Check scene coverage"""

    logger = logging.getLogger("Coverage")

    def __init__(self):
        super(CoverageCheck, self).__init__()

    def pre_invoke(self):
        """Pre-invoke"""
        pass

    def invoke(self, context):
        """Invoke"""
        self.use_lock = context.get("use_lock", False)
        self.logger.debug("Locking is used in coverage checker: %s",
                          str(self.use_lock))
        if self.use_lock:
            self.logger.debug("Scene loader acquires lock of previous "
                              "worker: %s", str(context["prev_lock"]))
            utils.acquire_lock(context["prev_lock"])

        scene = context["content"]
        overpass = Pass(scene.info["platform_name"],
                        scene.info['start_time'],
                        scene.info['end_time'],
                        instrument=scene.info["sensor"][0])
        areas = []
        for area_name in scene.info["areas"]:
            self.logger.info("Checking coverage of %s", area_name)

            try:
                min_coverage = context["min_coverage"][area_name]
            except KeyError:
                self.logger.warning("No minimum coverage given, "
                                    "assuming 0 % coverage needed")
                areas.append(area_name)
                continue

            if utils.covers(overpass, area_name, min_coverage, self.logger):
                areas.append(area_name)
            else:
                self.logger.info("Area coverage too low, skipping %s",
                                 area_name)
                continue

        if len(areas) > 0:
            scene.info["areas"] = areas
            context["output_queue"].put(scene)
        else:
            self.logger.info("No areas with enough coverage")

        if utils.release_locks([context["lock"]]):
            self.logger.debug("Scene loader releases own lock %s",
                              str(context["lock"]))
            time.sleep(1)

        # Wait until the lock has been released downstream
        if self.use_lock:
            utils.acquire_lock(context["lock"])
            utils.release_locks([context["lock"]])

        # After all the items have been processed, release the lock for
        # the previous step
        utils.release_locks([context["prev_lock"]], log=self.logger.debug,
                            log_msg="Scene loader releases lock of " +
                            "previous worker")

    def post_invoke(self):
        """Post-invoke"""
        pass



"""Classes for generating image composites for Trollflow based
Trollduction using satpy"""

import logging
import time

import yaml

from trollflow.workflow_component import AbstractWorkflowComponent
from trollflow_sat import utils
from trollsched.satpass import Pass


class Resampler(AbstractWorkflowComponent):

    """Creates resampled local area scenes."""

    logger = logging.getLogger("Resampler")

    def __init__(self):
        super(Resampler, self).__init__()

    def pre_invoke(self):
        """Pre-invoke"""
        pass

    def invoke(self, context):
        """Invoke"""
        # Set locking status, default to False
        self.use_lock = context.get("use_lock", False)
        self.logger.debug("Locking is used in resampler: %s",
                          str(self.use_lock))
        if self.use_lock:
            self.logger.debug("Compositor acquires lock of previous "
                              "worker: %s", str(context["prev_lock"]))
            utils.acquire_lock(context["prev_lock"])

        glbl = context["content"]
        with open(context["product_list"], "r") as fid:
            product_config = yaml.load(fid)

        # Handle config options
        kwargs = {}

        kwargs['precompute'] = context.get('precompute', False)
        kwargs['mask_area'] = context.get('mask_area', True)
        self.logger.debug("Setting precompute to %s and masking to %s",
                          str(kwargs['precompute']), str(kwargs['mask_area']))

        kwargs['nprocs'] = context.get('nprocs', 1)
        self.logger.debug("Using %d CPUs for resampling.", kwargs['nprocs'])

        kwargs['resampler'] = context.get('proj_method', "nearest")
        self.logger.debug("Using resampling method: '%s'.",
                          kwargs['resampler'])

        try:
            kwargs['cache_dir'] = context['cache_dir']
            self.logger.debug("Setting projection cache dir to %s",
                              kwargs['cache_dir'])
        except (AttributeError, KeyError):
            pass

        prod_list = product_config["product_list"]

        # Overpass for coverage calculations
        overpass = Pass(glbl.info['platform_name'],
                        glbl.info['start_time'],
                        glbl.info['end_time'],
                        instrument=glbl.info['sensor'][0])

        for area_id in prod_list:
            # Check for area coverage
            min_coverage = prod_list[area_id].get("min_coverage", 0.0)
            if not utils.covers(overpass, area_id, min_coverage, self.logger):
                continue

            kwargs['radius_of_influence'] = None
            try:
                area_config = product_config["product_list"][area_id]
                kwargs['radius_of_influence'] = \
                    area_config.get("srch_radius", context["radius"])
            except (AttributeError, KeyError):
                kwargs['radius_of_influence'] = 10000.

            if kwargs['radius_of_influence'] is None:
                self.logger.debug("Using default search radius.")
            else:
                self.logger.debug("Using search radius %d meters.",
                                  int(kwargs['radius_of_influence']))
            # Set lock if locking is used
            if self.use_lock:
                self.logger.debug("Resampler acquires own lock %s",
                                  str(context["lock"]))
                utils.acquire_lock(context["lock"])
            # if area_id not in glbl.info["areas"]:
            #     utils.release_locks([context["lock"]])
            #     continue

            if area_id == "satproj":
                self.logger.info("Using satellite projection")
                lcl = glbl
            else:
                try:
                    metadata = glbl.attrs
                except AttributeError:
                    metadata = glbl.info
                self.logger.info("Resampling time slot %s to area %s",
                                 metadata["start_time"], area_id)
                lcl = glbl.resample(area_id, **kwargs)
            try:
                metadata = lcl.attrs
            except AttributeError:
                metadata = lcl.info
            metadata["product_config"] = product_config
            metadata["area_id"] = area_id
            metadata["products"] = prod_list[area_id]['products']

            self.logger.debug("Inserting lcl (area: %s, start_time: %s) "
                              "to writer's queue",
                              area_id, str(metadata["start_time"]))
            context["output_queue"].put(lcl)
            del lcl
            lcl = None

            if utils.release_locks([context["lock"]]):
                self.logger.debug("Resampler releases own lock %s",
                                  str(context["lock"]))
                # Wait 1 second to ensure next worker has time to acquire the
                # lock
                time.sleep(1)

        # Wait until the lock has been released downstream
        if self.use_lock:
            utils.acquire_lock(context["lock"])
            utils.release_locks([context["lock"]])

        # After all the items have been processed, release the lock for
        # the previous step
        utils.release_locks([context["prev_lock"]], log=self.logger.debug,
                            log_msg="Resampler releses lock of previous " +
                            "worker: %s" % str(context["prev_lock"]))

    def post_invoke(self):
        """Post-invoke"""
        pass




import logging
import os.path
import Queue
import time
from threading import Thread

from posttroll.message import Message
from posttroll.publisher import Publish
from trollflow_sat import utils
from trollsift import compose


class DataWriterContainer(object):

    '''Container for DataWriter instance
    '''

    logger = logging.getLogger("DataWriterContainer")
    _prev_lock = None

    def __init__(self, topic=None, port=0, nameservers=None,
                 save_settings=None, use_lock=False,
                 publish_vars=None):
        # store parameters for later writer restarts
        self.topic = topic
        self._input_queue = None
        self.output_queue = None  # Queue.Queue()
        self.thread = None
        self.use_lock = use_lock
        self._save_settings = save_settings
        self._topic = topic
        self._port = port
        self._nameservers = nameservers
        self._publish_vars = publish_vars
        self._init_writer()

    def _init_writer(self):
        # Create a Writer instance
        self.writer = DataWriter(queue=self.input_queue,
                                 save_settings=self._save_settings,
                                 topic=self._topic,
                                 port=self._port,
                                 nameservers=self._nameservers,
                                 publish_vars=self._publish_vars,
                                 prev_lock=self._prev_lock)
        # Start Writer instance into a new daemonized thread.
        self.thread = Thread(target=self.writer.run)
        self.thread.setDaemon(True)
        self.thread.start()

    @property
    def input_queue(self):
        """Property writer"""
        return self._input_queue

    @input_queue.setter
    def input_queue(self, queue):
        """Set writer queue"""
        self._input_queue = queue
        self.writer.queue = queue

    @property
    def prev_lock(self):
        """Property writer"""
        return self._prev_lock

    @prev_lock.setter
    def prev_lock(self, lock):
        """Set lock of the previous worker"""
        if self.use_lock:
            self._prev_lock = lock
            self.writer.prev_lock = lock

    def __setstate__(self, state):
        self.__init__(**state)

    def restart(self):
        '''Restart writer after configuration update.
        '''
        if self.writer is not None:
            if self.writer.loop:
                self.stop()
        self._init_writer()

    def stop(self):
        '''Stop writer.'''
        self.logger.debug("Stopping writer.")
        self.writer.stop()
        self.thread.join()
        self.logger.debug("Writer stopped.")
        self.thread = None

    def is_alive(self):
        """Return the thread status"""
        return self.thread.is_alive()


class DataWriter(Thread):

    """Writes data to disk.
    """

    logger = logging.getLogger("DataWriter")

    def __init__(self, queue=None, save_settings=None,
                 topic=None, port=0, nameservers=None, prev_lock=None,
                 publish_vars=None):
        Thread.__init__(self)
        self.queue = queue
        self._loop = False
        self._save_settings = save_settings
        self._port = port
        if nameservers is None:
            self._nameservers = []
        else:
            if type(nameservers) not in (list, tuple, set):
                nameservers = [nameservers, ]
            self._nameservers = nameservers
        self._topic = topic
        self.prev_lock = prev_lock
        self._publish_vars = publish_vars or {}

    def run(self):
        """Run the thread."""
        self._loop = True
        # Parse settings for saving
        compression = self._save_settings.get('compression', 6)
        tags = self._save_settings.get('tags', None)
        fformat = self._save_settings.get('fformat', None)
        gdal_options = self._save_settings.get('gdal_options', None)
        blocksize = self._save_settings.get('blocksize', None)

        kwargs = {'compression': compression,
                  'tags': tags,
                  'fformat': fformat,
                  'gdal_options': gdal_options,
                  'blocksize': blocksize}

        # Initialize publisher context
        with Publish("l2producer", port=self._port,
                     nameservers=self._nameservers) as pub:

            while self._loop:
                if self.queue is not None:
                    try:
                        lcl = self.queue.get(True, 1)
                        if self.prev_lock is not None:
                            self.logger.debug("Writer acquires lock of "
                                              "previous worker: %s",
                                              str(self.prev_lock))
                            utils.acquire_lock(self.prev_lock)
                        self.queue.task_done()
                    except Queue.Empty:
                        # After all the items have been processed, release the
                        # lock for the previous worker
                        continue

                    try:
                        info = lcl.attrs.copy()
                        product_config = lcl.attrs["product_config"]
                        products = lcl.attrs["products"]
                    except AttributeError:
                        info = lcl.info.copy()
                        product_config = lcl.info["product_config"]
                        products = lcl.info["products"]

                    # Available composite names
                    composite_names = [dset.name for dset in lcl.keys()]

                    for i, prod in enumerate(products):
                        # Skip the removed composites
                        if prod not in composite_names:
                            continue
                        fnames, _ = utils.create_fnames(info,
                                                        product_config,
                                                        prod)
                        # Some of the files might have specific
                        # writers, use them if configured
                        writers = utils.get_writer_names(product_config, prod,
                                                         info["area_id"])

                        for j, fname in enumerate(fnames):
                            if writers[j]:
                                self.logger.info("Saving %s with writer %s",
                                                 fname, writers[j])
                            else:
                                self.logger.info(
                                    "Saving %s with default writer", fname)

                            lcl.save_dataset(prod,
                                             filename=fname,
                                             writer=writers[j],
                                             **kwargs)

                            self.logger.info("Saved %s", fname)

                            try:
                                area = lcl[prod].attrs.get("area")
                            except AttributeError:
                                area = lcl[prod].info.get("area")

                            try:
                                area_data = {"name": area.name,
                                             "area_id": area.area_id,
                                             "proj_id": area.proj_id,
                                             "proj4": area.proj4_string,
                                             "shape": (area.x_size,
                                                       area.y_size)
                                             }
                            except AttributeError:
                                area_data = None

                            to_send = dict(info) if '*' \
                                in self._publish_vars else {}

                            for dest_key in self._publish_vars:
                                if dest_key != '*':
                                    to_send[dest_key] = info.get(
                                        self._publish_vars[dest_key]
                                        if
                                        isinstance(self._publish_vars, dict)
                                        else
                                        dest_key)

                            to_send_fix = {"nominal_time": info["start_time"],
                                           "uid": os.path.basename(fname),
                                           "uri": os.path.abspath(fname),
                                           "area": area_data,
                                           "productname": info["productname"]
                                           }
                            to_send.update(to_send_fix)

                            if self._topic is not None:
                                topic = self._topic
                                if area_data is not None:
                                    topic = compose(topic,  area_data)
                                else:
                                    topic = compose(topic,
                                                    {'area_id': 'satproj'})

                                msg = Message(topic, "file", to_send)
                                pub.send(str(msg))
                                self.logger.debug("Sent message: %s", str(msg))

                    del lcl
                    lcl = None
                    # After all the items have been processed, release the
                    # lock for the previous worker
                    if self.prev_lock is not None:
                        utils.release_locks([self.prev_lock],
                                            self.logger.debug,
                                            "Writer releses lock of "
                                            "previous worker: %s" %
                                            str(self.prev_lock))
                else:
                    time.sleep(1)

    def stop(self):
        """Stop writer."""
        self._loop = False

    @property
    def loop(self):
        """Property loop"""
        return self._loop



"""Classes for handling file fetching for Trollflow based Trollduction"""

import logging
import os
import socket
import time
from tempfile import gettempdir
from urlparse import urlparse, urlsplit

import netifaces

from posttroll.message import Message
from trollflow.utils import acquire_lock, release_lock
from trollflow.workflow_component import AbstractWorkflowComponent

logger = logging.getLogger(__name__)


def get_local_ips():
    inet_addrs = [netifaces.ifaddresses(iface).get(netifaces.AF_INET)
                  for iface in netifaces.interfaces()]
    ips = []
    for addr in inet_addrs:
        if addr is not None:
            for add in addr:
                ips.append(add['addr'])
    return ips


def is_uri_on_server(uri, strict=False):
    """Check if the *uri* is designating a place on the server.

    If *strict* is True, the hostname has to be specified in the *uri* for the
    path to be considered valid.
    """
    url = urlparse(uri)
    try:
        url_ip = socket.gethostbyname(url.hostname)
    except (socket.gaierror, TypeError):
        if strict:
            return False
        try:
            os.stat(url.path)
        except OSError:
            return False
    else:
        if url.hostname == '':
            if strict:
                return False
            try:
                os.stat(url.path)
            except OSError:
                return False
        elif url_ip not in get_local_ips():
            return False
        else:
            try:
                os.stat(url.path)
            except OSError:
                return False
    return True


def check_uri(uri):
    """Check that the provided *uri* is on the local host and return the
    file path.
    """
    if isinstance(uri, (list, set, tuple)):
        paths = [check_uri(ressource) for ressource in uri]
        return paths
    url = urlparse(uri)
    try:
        if url.hostname:
            url_ip = socket.gethostbyname(url.hostname)

            if url_ip not in get_local_ips():
                try:
                    os.stat(url.path)
                except OSError:
                    raise IOError(
                        "Data file %s unaccessible from this host" % uri)

    except socket.gaierror:
        logger.warning("Couldn't check file location, running anyway")

    return url.path


class Fetcher(AbstractWorkflowComponent):

    """Fetch files from the network if needed, and sanitize the pathname."""

    def __init__(self):
        super(Fetcher, self).__init__()

    def pre_invoke(self):
        """Pre-invoke."""
        pass

    def invoke(self, context):
        """Invoke."""
        # Set locking status, default to False
        self.use_lock = context.get("use_lock", False)
        logger.debug("Locking is used in Fetcher: %s",
                     str(self.use_lock))
        if self.use_lock:
            logger.debug("Fetcher acquires lock of previous "
                         "worker: %s", str(context["prev_lock"]))
            acquire_lock(context["prev_lock"])

        logger.info("Fetching files.")
        message = fetch_files(context["content"], context.get("destination",
                                                              gettempdir()))
        context["output_queue"].put(message)

        if self.use_lock:
            logger.debug("Fetcher releases own lock %s",
                         str(context["lock"]))
            release_lock(context["lock"])
            # Wait 1 second to ensure next worker has time to acquire the
            # lock
            time.sleep(1)

        # Wait until the lock has been released downstream
        if self.use_lock:
            acquire_lock(context["lock"])
            release_lock(context["lock"])

        # After all the items have been processed, release the lock for
        # the previous step
        logger.debug("Fetcher releases lock of previous worker")
        release_lock(context["prev_lock"])

    def post_invoke(self):
        """Post-invoke"""
        pass


def fetch_files(message, destination):
    """Fetch files from the network"""
    new_message = Message(rawstr=str(message))
    if new_message.type == "dataset":
        for dset in new_message.data["dataset"]:
            dset["uri"] = fetch_file(dset["uri"], destination)
    elif new_message.type == "collection":
        raise NotImplementedError
    else:
        new_message.data["uri"] = fetch_file(new_message.data["uri"],
                                             destination)
    return new_message


def fetch_file(uri, destination):
    """Fetch a single file into `destination` and return its pathname."""
    try:
        return check_uri(uri)
    except IOError:
        del destination
        raise NotImplementedError("Don't know how to fetch non local files")



import logging
import os.path

from posttroll.message import Message
from posttroll.publisher import Publish
from trollflow.utils import acquire_lock as trollflow_acquire_lock
from trollflow.utils import release_lock
from trollsift import compose
from trollsift.parser import _extract_parsedef as extract_parsedef
try:
    from satpy.resample import get_area_def
except ImportError:
    from mpop.projector import get_area_def

try:
    from pyorbital import astronomy
except ImportError:
    astronomy = None

PATTERN = "{time:%Y%m%d_%H%M}_{platform_name}_{areaname}_{productname}.png"
FORMAT = "png"

LOGGER = logging.getLogger(__name__)


def get_prerequisites_yaml(global_data, prod_list, area_list):
    """Get composite prerequisite channels for a list of areas"""
    reqs = set()
    for area in area_list:
        for prod_id in prod_list[area]["products"]:
            try:
                composite = getattr(global_data.image, prod_id)
            except AttributeError:
                continue
            reqs |= composite.prerequisites
    return reqs


def create_fnames(info, product_config, prod_id):
    """Create filename for product *prod*"""
    area_id = info["area_id"]

    # List of products
    products = product_config["product_list"][area_id]["products"]

    # Get area name
    info["areaname"] = product_config["product_list"][area_id]["areaname"]

    # Find output directory
    output_dir = products[prod_id].get("output_dir", "")
    if output_dir == "":
        output_dir = product_config["common"].get("output_dir", "")

    if output_dir == "":
        LOGGER.warning("No output directory specified, "
                       "saving to current directory!")

    # Find filename pattern
    pattern = products[prod_id].get("fname_pattern", "")
    if pattern == "":
        pattern = product_config["common"].get("fname_pattern", "")

    if pattern == "":
        LOGGER.warning("No pattern was given, using built-in default: %s",
                       PATTERN)
        pattern = PATTERN

    # Join output dir and filename pattern
    pattern = os.path.join(output_dir, pattern)

    # Find output formats
    formats = products[prod_id].get("formats", ["", ])
    if formats[0] == "":
        formats = product_config["common"].get("formats", [{"format": FORMAT,
                                                            "writer": None}])

    prod_name = products[prod_id]["productname"]
    info["productname"] = prod_name

    # Find the name of the available 'nominal_time'
    time_name = None
    for key in info:
        if "time" in key and "end" not in key and "proc" not in key:
            time_name = key
            LOGGER.debug("metadata time name is '%s'", time_name)
            break

    if time_name is None and "time" in pattern:
        return None

    # Adjust filename pattern so that time_name is present.
    # Get parse definitions and try to figure out if there's
    # an item for time
    parsedefs, _ = extract_parsedef(pattern)
    for itm in parsedefs:
        if isinstance(itm, dict):
            key, val = itm.items()[0]
            if val is None:
                continue
            # Need to exclude 'end_time' and 'proc_time' / 'processing_time'
            if ("time" in key or "%" in val) and \
               "end" not in key and key != time_name:
                LOGGER.debug("Updating pattern from '%s' ...", pattern)

                while '{' + key in pattern:
                    pattern = pattern.replace('{' + key,
                                              '{' + time_name)
                LOGGER.debug("... to '%s'", pattern)

    fnames = []
    for fmt in formats:
        info["format"] = fmt["format"]
        # Ensure non-unicode filename
        fnames.append(str(compose(pattern, info)))

    return (fnames, prod_name)


def get_writer_names(product_config, prod_id, area_id):
    """Get writer names for the """
    products = product_config["product_list"][area_id]["products"]
    formats = products[prod_id].get("formats", ["", ])
    if formats[0] == "":
        formats = product_config["common"].get("formats", [{"format": FORMAT,
                                                            "writer": None}])
    writers = []
    for fmt in formats:
        writers.append(fmt.get("writer", None))

    return writers


def get_satpy_group_composite_names(product_config, group):
    """Parse composite names from the product config for the given
    group."""
    composites = set()
    prod_list = product_config['product_list']
    for group in product_config['groups'][group]:
        composites.update(set(prod_list[group]['products'].keys()))
    return composites


def get_satpy_area_composite_names(product_config, area_id):
    """Parse composite names from the product config for the given
    group."""
    prod_list = product_config['product_list']
    return prod_list[area_id]['products'].keys()


def find_time_name(info):
    """Try to find the name for 'nominal' time"""
    for key in info:
        if "time" in key and "end" not in key and "proc" not in key:
            return key
    return None


def bad_sunzen_range_satpy(product_config, group, composite, start_time):
    """Check if Sun zenith angle is valid at the configured location.
    SatPy version.
    TODO: refactor with bad_sunzen_range()
    """
    # FIXME: check all areas within the group
    area_id = product_config['groups'][group][0]
    product_conf = \
        product_config["product_list"][area_id]["products"][composite]

    if ("sunzen_night_minimum" not in product_conf and
            "sunzen_day_maximum" not in product_conf):
        return False

    if astronomy is None:
        LOGGER.warning("Pyorbital not installed, unable to calculate "
                       "Sun zenith angles!")
        return False

    if "sunzen_lon" not in product_conf and "sunzen_lat" not in product_conf:
        LOGGER.warning("No 'sunzen_lon' or 'sunzen_lat' configured, "
                       "can\'t check Sun elevation.")
        return False

    lon = product_conf["sunzen_lon"]
    lat = product_conf["sunzen_lat"]
    sunzen = astronomy.sun_zenith_angle(start_time, lon, lat)
    LOGGER.debug("Sun zenith angle is %.2f degrees", sunzen)

    try:
        limit = product_conf["sunzen_night_minimum"]
        if sunzen < limit:
            return True
        else:
            return False
    except KeyError:
        pass

    try:
        limit = product_conf["sunzen_day_maximum"]
        if sunzen > limit:
            return True
        else:
            return False
    except KeyError:
        pass


def bad_sunzen_range(area, product_config, area_id, prod, time_slot):
    """Check if Sun zenith angle is valid at the configured location."""
    product_conf = product_config["product_list"][area_id]["products"][prod]

    if ("sunzen_night_minimum" not in product_conf and
            "sunzen_day_maximum" not in product_conf):
        return False

    if astronomy is None:
        LOGGER.warning("Pyorbital not installed, unable to calculate "
                       "Sun zenith angles!")
        return False

    if area.lons is None:
        area.lons, area.lats = area.get_lonlats()

    lon, lat = None, None

    try:
        lon = product_conf["sunzen_lon"]
        lat = product_conf["sunzen_lat"]
    except KeyError:
        pass

    if lon is None or lat is None:
        try:
            x_idx = product_conf["sunzen_x_idx"]
            y_idx = product_conf["sunzen_y_idx"]
            lon = area.lons[x_idx]
            lat = area.lats[y_idx]
        except KeyError:
            pass

    if lon is None or lat is None:
        LOGGER.info("Using area center for Sun zenith angle calculation")
        y_idx = int(area.y_size / 2)
        x_idx = int(area.x_size / 2)
        lon, lat = area.get_lonlat(y_idx, x_idx)

    sunzen = astronomy.sun_zenith_angle(time_slot, lon, lat)
    LOGGER.debug("Sun zenith angle is %.2f degrees", sunzen)

    try:
        limit = product_conf["sunzen_night_minimum"]
        if sunzen < limit:
            return True
        else:
            return False
    except KeyError:
        pass

    try:
        limit = product_conf["sunzen_day_maximum"]
        if sunzen > limit:
            return True
        else:
            return False
    except KeyError:
        pass


def send_message(topic, msg_type, msg_data, nameservers=None, port=0):
    """Send monitoring message"""
    if nameservers is None:
        nameservers = []
    if not isinstance(nameservers, list):
        nameservers = [nameservers]

    with Publish("trollflow-sat", port=port,
                 nameservers=nameservers) as pub:
        msg = Message(topic, msg_type, msg_data)
        pub.send(str(msg))


def _get_data_time_from_message_data(msg_data):
    """Get data timestamp from message data"""
    if "nominal_time" in msg_data:
        data_time = msg_data["nominal_time"]
    elif "start_time" in msg_data:
        data_time = msg_data["start_time"]
    elif "time" in msg_data:
        data_time = msg_data["time"]
    else:
        data_time = None

    return data_time


def _get_orbit_number_from_message_data(msg_data):
    """Get orbit number from message data"""
    if "orbit_number" in msg_data:
        return msg_data["orbit_number"]
    else:
        return None


def get_monitor_metadata(msg, status=None, service=None):
    """Collect metadata for monitoring message"""
    data = {"message_time": msg.time,
            "data_time": _get_data_time_from_message_data(msg.data),
            "platform_name": msg.data["platform_name"],
            "sensor": msg.data["sensor"],
            "orbit_number": _get_orbit_number_from_message_data(msg.data),
            "status": status,
            "service": service}

    return data


def release_locks(locks, log=None, log_msg=None):
    """Release locks and optionnally send log message to *log* function"""
    if not isinstance(locks, list):
        locks = [locks]
    if log is not None and log_msg is not None:
        log(log_msg)
    ret_vals = []
    for lock in locks:
        ret_vals.append(release_lock(lock))

    return max(ret_vals)


def acquire_lock(lock):
    """Acquire the given lock"""
    return trollflow_acquire_lock(lock)


def covers(overpass, area_name, min_coverage, logger):
    try:
        area_def = get_area_def(area_name)
        if min_coverage == 0 or overpass is None:
            return True
        min_coverage /= 100.0
        coverage = overpass.area_coverage(area_def)
        if coverage <= min_coverage:
            logger.info("Coverage too small %.1f%% (out of %.1f%%) "
                        "with %s",
                        coverage * 100, min_coverage * 100,
                        area_name)
            return False
        else:
            logger.info("Coverage %.1f%% with %s",
                        coverage * 100, area_name)

    except AttributeError:
        logger.warning("Can't compute area coverage with %s!",
                       area_name)
    return True
